def pytest_configure():
    from django.conf import settings

    settings.configure(
        DEBUG_PROPAGATE_EXCEPTIONS=True,
        DATABASES={
            "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}
        },
        SITE_ID=1,
        SECRET_KEY="fake-key",
        INSTALLED_APPS=(
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django_und",
            "tests",
        ),
    )



from django.test import TestCase

from django_factory_boy import auth as auth_factories

from tests.models import Article, create_article


class TestVote(TestCase):
    def setUp(self):
        self.user = auth_factories.UserFactory()
        self.user.active = True
        self.user2 = auth_factories.UserFactory()
        self.user2.active = True

        self.article = create_article("Some title", self.user)
        self.article2 = create_article("Some title", self.user2)

    @staticmethod
    def _assert_article_score(article, score, up, down):
        assert article.und_score == score
        assert article.und_score_up == up
        assert article.und_score_down == down

        article_db = Article.objects.get(pk=article.pk)
        assert article_db.und_score == score
        assert article_db.und_score_up == up
        assert article_db.und_score_down == down

    def test_model(self):
        self.article.upvote(self.user)
        assert len(self.article.und_votes.all()) == 1
        vote = self.article.und_votes.all()[0]
        assert str(vote) == "{}:{}:{}".format(self.user, self.article, 1)

    def test_model_username(self):
        with self.settings(UND_USE_USERNAME=True):
            self.article.upvote(self.user.username)
            assert len(self.article.und_votes.all()) == 1
            vote = self.article.und_votes.all()[0]
            assert str(vote) == "{}:{}:{}".format(self.user.username, self.article, 1)

    def test_upvote(self):
        self.article.upvote(self.user)
        self._assert_article_score(self.article, 1, 1, 0)

    def test_downvote(self):
        self.article.downvote(self.user)
        self._assert_article_score(self.article, -1, 0, -1)

    def test_cancel_upvote(self):
        self.article.upvote(self.user)
        self.article.upvote(self.user)
        self._assert_article_score(self.article, 0, 0, 0)

    def test_cancel_downvote(self):
        self.article.downvote(self.user)
        self.article.downvote(self.user)
        self._assert_article_score(self.article, 0, 0, 0)

    def test_upvote_then_downvote(self):
        self.article.upvote(self.user)
        self.article.downvote(self.user)
        self._assert_article_score(self.article, -1, 0, -1)

    def test_downvote_then_upvote(self):
        self.article.downvote(self.user)
        self.article.upvote(self.user)
        self._assert_article_score(self.article, 1, 1, 0)

    def test_multiple_votes(self):
        self.article.upvote(self.user)
        self.article.downvote(self.user2)
        self._assert_article_score(self.article, 0, 1, -1)

        self.article.upvote(self.user2)
        self._assert_article_score(self.article, 2, 2, 0)

        self.article.downvote(self.user)
        self._assert_article_score(self.article, 0, 1, -1)

        self.article.downvote(self.user2)
        self._assert_article_score(self.article, -2, 0, -2)

    def test_reset_score(self):
        self.article.upvote(self.user)
        self.article.downvote(self.user2)
        self.article.reset_und_score()
        self._assert_article_score(self.article, 0, 1, -1)



"""Vote and VoteMixin."""
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Sum, F
from django.utils.translation import ugettext_lazy as _


class Vote(models.Model):
    """
    Up and down vote model.

    By default this app use a foreignkey to a django user
    it also provide a method to use a username instead of a django
    user this allows the model to be used with various authentification
    policies.
    """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        null=True,
        related_name="und_votes",
    )
    username = models.CharField(max_length=255, null=True)
    score = models.IntegerField(default=1)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    class Meta:
        verbose_name = _("Vote")
        verbose_name_plural = _("Votes")

    def __str__(self):
        """Return user, object, and score."""
        if getattr(settings, "UND_USE_USERNAME", False):
            return "{}:{}:{}".format(self.username, self.content_object, self.score)
        return "{}:{}:{}".format(self.user, self.content_object, self.score)

    @classmethod
    def vote_with_user_object(cls, content_object, user_object, score):
        """
        Return a Vote instance.

        Return a Vote instance for a user, a voted object and a score.
        """
        if getattr(settings, "UND_USE_USERNAME", False):
            return Vote(
                content_object=content_object, username=user_object, score=score
            )
        return Vote(content_object=content_object, user=user_object, score=score)


class VoteMixin(models.Model):
    """Mixin to attach to a model that has up and down votes."""

    und_votes = GenericRelation(Vote)
    und_score_up = models.IntegerField(default=0)
    und_score_down = models.IntegerField(default=0)

    class Meta:
        abstract = True

    @property
    def und_score(self):
        """Return the model score."""
        return self.und_score_up + self.und_score_down

    def _get_with_object(self, user_object):
        """Return a vote based on object and settings (UND_USE_USERNAME)."""
        if getattr(settings, "UND_USE_USERNAME", False):
            return self.und_votes.get(username=user_object)
        return self.und_votes.get(user=user_object)

    def upvote(self, user_object):
        """Upvote the instance with provided user."""
        # Change of score_up
        diff_up = 0
        # Change of score down
        diff_down = 0

        try:
            # Already voted content
            vote = self._get_with_object(user_object)
            if vote.score == 1:
                # Cancel previous upvote
                vote.delete()
                # Remove 1 upvote
                diff_up = -1
            else:
                # Previously downvoted
                vote.score = 1
                vote.save()
                # Remove downvote and add upvote
                diff_down = 1
                diff_up = 1
        except Vote.DoesNotExist:
            vote = Vote.vote_with_user_object(
                content_object=self, user_object=user_object, score=1
            )
            vote.save()
            # Create an upvote
            diff_up = 1
        self.und_score_up += diff_up
        self.und_score_down += diff_down

        # Update self score, use update and filter to avoid triggering signals
        self.__class__.objects.filter(id=self.id).update(
            und_score_up=F("und_score_up") + diff_up,
            und_score_down=F("und_score_down") + diff_down,
        )

    def downvote(self, user_object):
        """Downvote the instance with the provided user."""
        # Change of score_up
        diff_up = 0
        # Change of score down
        diff_down = 0

        try:
            # Already voted content
            vote = self._get_with_object(user_object)
            if vote.score == -1:
                # Cancel previous downvote
                vote.delete()
                # Remove 1 downvote
                diff_down = 1
            else:
                # Previously upvoted
                vote.score = -1
                vote.save()
                # Remove upvote and add downvote
                diff_up = -1
                diff_down = -1
        except Vote.DoesNotExist:
            vote = Vote.vote_with_user_object(
                content_object=self, user_object=user_object, score=-1
            )
            vote.save()
            # Create downvote
            diff_down = -1

        self.und_score_up += diff_up
        self.und_score_down += diff_down
        # Update self score, use update and filter to avoid triggering signals
        self.__class__.objects.filter(id=self.id).update(
            und_score_up=F("und_score_up") + diff_up,
            und_score_down=F("und_score_down") + diff_down,
        )

    def reset_und_score(self):
        """Reset score to the correct count (should not be necessary)."""
        score_up = (
            self.und_votes.filter(score__gt=0).aggregate(Sum("score"))["score__sum"]
            or 0
        )
        score_down = (
            self.und_votes.filter(score__lt=0).aggregate(Sum("score"))["score__sum"]
            or 0
        )
        self.und_score_up = score_up
        self.und_score_down = score_down
        # Update self score, use update and filter to avoid triggering signals
        self.__class__.objects.filter(id=self.id).update(
            und_score_up=score_up, und_score_down=score_down
        )



from distutils.core import setup
from setuptools import find_packages

__version__ = "0.1.1"

install_requires = ["django"]

tests_require = [
    "pytest",
    "pytest-cov",
    "pytest-django",
    "django_factory_boy",
    "codecov",
]

extras = {"test": tests_require}


setup(
    name="django-und",
    packages=find_packages(),
    version=__version__,
    description=(
        "Django Up and Down, django app " "to add upvote and downvote to models"
    ),
    author="luxcem",
    author_email="a@luxcem.fr",
    url="https://github.com/luxcem/django-und",
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras,
    setup_requires=["pytest-runner"],
)



from django.contrib.auth.models import User
from django.db import models

from django_und.models import VoteMixin


def create_article(title, author, collaborators=None):
    if not collaborators:
        collaborators = []

    article = Article.objects.create(title=title, content=title * 20, author=author)
    article.save()
    if collaborators:
        article.collaborators.add(*collaborators)

    return article


class Article(VoteMixin, models.Model):
    title = models.CharField("title", max_length=200)
    content = models.TextField("content")
    author = models.ForeignKey(User, on_delete=models.CASCADE)
    collaborators = models.ManyToManyField(User)

    class Meta:
        app_label = "tests"

    def get_score(self):
        return self.und_score
